# Copyright (C) 2023 Associated Universities, Inc. Washington DC, USA.
#
# This file is part of NRAO Workspaces.
#
# Workspaces is free software: you can redistribute it and/or modify
# it under the terms of the GNU General Public License as published by
# the Free Software Foundation, either version 3 of the License, or
# (at your option) any later version.
#
# Workspaces is distributed in the hope that it will be useful,
# but WITHOUT ANY WARRANTY; without even the implied warranty of
# MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.  See the
# GNU General Public License for more details.
#
# You should have received a copy of the GNU General Public License
# along with Workspaces.  If not, see <https://www.gnu.org/licenses/>.
#
import shutil
from itertools import chain
from pathlib import Path
from typing import Callable, List, Optional

import pytest
import ws_annihilator
from requests_mock import Mocker
from sqlalchemy.util import itertools_filterfalse
from ws_annihilator import annihilator
from ws_annihilator.annihilator import (
    FAILURE_NOTIFICATION_TEMPLATE_NAME,
    Annihilator,
    CacheAnnihilator,
    Settings,
    SpoolAnnihilator,
    _arg_parser,
    run_annihilator,
)

MOCKED_SERVER_URL = "http://nonexistent-host"
MOCKED_NOTIFICATION_PORT = 3458
MOCKED_WORKFLOW_PORT = 3456
MOCKED_WORKFLOW_URL = f"{MOCKED_SERVER_URL}:{MOCKED_WORKFLOW_PORT}"
MOCKED_NOTIFICATION_URL = f"{MOCKED_SERVER_URL}:{MOCKED_NOTIFICATION_PORT}"


def make_mocked_rmtree(dirs_with_bad_permissions: List[Path]) -> Callable[[str], None]:
    """Make a replacement for shutil.rmtree that just raises a PermissionError when called on certain directories

    :param dirs_with_bad_permissions: Absolute paths of the directories to raise a PermissionError on
    :return: The mocked replacement for rmtree
    """
    dir_names_with_bad_permissions = [dir_path.name for dir_path in dirs_with_bad_permissions]

    def mocked_rmtree(path: str):
        if path in dir_names_with_bad_permissions:
            raise PermissionError()

    return mocked_rmtree


def requests_mock_notify_good(requests_mock: Mocker) -> Mocker:
    """Mock requests assuming notifications get sent successfully"""
    url = f"{MOCKED_NOTIFICATION_URL}/notify/ws_annihilator_failure/send"
    requests_mock.post(
        url,
        status_code=200,
        json={"message": "Email sent", "errors": {}},
        headers={
            "Content-Length": "39",
            "Content-Type": "application/json",
            "Date": "Mon, 06 Feb 2023 18:06:10 GMT",
            "Server": "waitress",
        },
    )
    return requests_mock


def requests_mock_spool_annihilator(
    requests_mock: Mocker, annihilator: SpoolAnnihilator, stale_dirs: List[str]
) -> Mocker:
    """Mock the two endpoints that a SpoolAnnihilator needs to access"""
    requests_mock.get(
        annihilator.stale_directory_url,
        status_code=200,
        json={"stale_dirs": ",".join(stale_dirs)},
    )
    requests_mock.put(annihilator.clean_url, status_code=200)
    return requests_mock


def setup_dir_with_file(tmp_path: Path, dir_name: str) -> Path:
    dir_with_permissions_issue = tmp_path / dir_name
    dir_with_permissions_issue.mkdir()
    failed_file = dir_with_permissions_issue / "failure.txt"
    failed_file.touch(exist_ok=False)
    return dir_with_permissions_issue


def make_settings_for_dir(dir: Path) -> Settings:
    days = str(10)
    return {
        "notification_url": MOCKED_NOTIFICATION_URL,
        "failure_notification_email": "nonexistent@email.address",
        "workflow_url": MOCKED_WORKFLOW_URL,
        "cache": str(dir),
        "spool": str(dir),
        "storage": str(dir),
        "staging": str(dir),
        "spool_days": days,
        "cache_days": days,
        "staging_days": days,
        "storage_days": days,
        "calibration_days": days,
        "image_days": days,
    }


def test_annihilate_directory_failure(tmp_path: Path, requests_mock: Mocker, monkeypatch: pytest.MonkeyPatch):
    args = _arg_parser().parse_args(["--profile", "nonexistent", "--all"])
    dir_to_annihilate = setup_dir_with_file(tmp_path, "tmp12345")
    annihilator = SpoolAnnihilator(args, make_settings_for_dir(dir_to_annihilate))
    requests_mock = requests_mock_spool_annihilator(requests_mock, annihilator, [str(dir_to_annihilate)])
    # Mock rmtree to force a PermissionError, causing annihilate_directory() to fail
    monkeypatch.setattr("shutil.rmtree", make_mocked_rmtree([dir_to_annihilate]))

    did_fail = SpoolAnnihilator(args, make_settings_for_dir(dir_to_annihilate)).annihilate_directory(
        str(dir_to_annihilate)
    )
    assert did_fail is True
    # The filesystem won't have changed since rmtree is mocked, so no need to assert about it


def test_annihilate_directory_success(tmp_path: Path, requests_mock: Mocker):
    args = _arg_parser().parse_args(["--profile", "nonexistent", "--all"])
    dir_to_annihilate = setup_dir_with_file(tmp_path, "tmp12345")
    assert dir_to_annihilate.exists()

    annihilator = SpoolAnnihilator(args, make_settings_for_dir(dir_to_annihilate))
    requests_mock = requests_mock_spool_annihilator(requests_mock, annihilator, [str(dir_to_annihilate)])
    did_succeed = annihilator.annihilate_directory(str(dir_to_annihilate))
    assert did_succeed is True
    assert not dir_to_annihilate.exists()


def test_multiple_failures(tmp_path: Path, requests_mock: Mocker, monkeypatch: pytest.MonkeyPatch):
    expected_failed_dirs = [setup_dir_with_file(tmp_path, f"failure-{i}") for i in range(3)]
    expected_successful_dir = setup_dir_with_file(tmp_path, "success")
    all_dirs = list(chain(expected_failed_dirs, [expected_successful_dir]))
    assert all(dir.exists() for dir in all_dirs)

    # Need to mock rmtree to get it to raise a PermissionError, triggering the notification
    monkeypatch.setattr("shutil.rmtree", make_mocked_rmtree(expected_failed_dirs))

    args = _arg_parser().parse_args(["--profile", "nonexistent"])  # Set up for spool annihilation
    settings = make_settings_for_dir(tmp_path)

    requests_mock = requests_mock_notify_good(requests_mock)
    # Needed for requests mocking since it knows what endpoints any instance with the same args and settings need to access
    annihilator = SpoolAnnihilator(args, settings)
    requests_mock = requests_mock_spool_annihilator(
        requests_mock, annihilator, [str(dir_path) for dir_path in all_dirs]
    )
    run_annihilator(args, settings)

    assert requests_mock.call_count > 1
    assert requests_mock.last_request
    notification_request = requests_mock.last_request
    assert notification_request.port == MOCKED_NOTIFICATION_PORT
    assert FAILURE_NOTIFICATION_TEMPLATE_NAME in notification_request.path
    assert notification_request.json().get("failed_directories") is not None
    assert set(notification_request.json()["failed_directories"]) == set(
        str(dir_path) for dir_path in expected_failed_dirs
    )
    # The filesystem won't have changed since rmtree is mocked, so no need to assert about it
